"""
Maintenance Impact Calculator support Endpoints
=================================================

These endpoints are intended for use by the Maintenance Impact Calculator tool

.. contents:: :local:

/mic/sites
----------------------

.. autofunction:: inventory_provider.routes.mic.sites


"""
import logging

from flask import Blueprint, request, Response

from inventory_provider.routes import common
from inventory_provider.routes.common import _ignore_cache_or_retrieve

logger = logging.getLogger(__name__)
routes = Blueprint('mic-support-routes', __name__)

ALL_DATA_SCHEMA = {
    "schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "patternProperties": {
        "^.*$": {
            "type": "object",
            "patternProperties": {
                "^.*$": {
                    "type": "object",
                    "patternProperties": {
                        "^.*$": {
                            "type": "array",
                            "items": {
                                "type": "object",
                                "properties": {
                                    "id": {"type": "integer"},
                                    "sid": {"type": "string"},
                                    "status": {"type": "string"},
                                    "name": {"type": "string"},
                                    "service_type": {"type": "string"},
                                    "contacts": {
                                        "type": "array",
                                        "items": {"type": "string"}
                                    },
                                    "planned_work_contacts": {
                                        "type": "array",
                                        "items": {"type": "string"}
                                    },
                                    "third_party_id": {"type": "string"}
                                },
                                "required": ["id", "sid", "status", "name",
                                             "service_type", "contacts",
                                             "planned_work_contacts"],  # noqa E501
                                "additionalProperties": False
                            }
                        }
                    }
                }
            }
        }
    },
    "additionalProperties": False
}

THIRD_PARTY_RESPONSE_SCHEMA = {
    "schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "properties": {
        "circuit_hierarchy": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "id": {"type": "integer"},
                    "status": {"type": "string"},
                    "name": {"type": "string"},
                    "service_type": {"type": "string"},
                    "contacts": {
                        "type": "array",
                        "items": {"type": "string"}
                    },
                    "planned_work_contacts": {
                        "type": "array",
                        "items": {"type": "string"}
                    },
                    "third_party_id": {"type": "string"}
                },
                "required": ["id", "status", "name", "service_type", "contacts", "planned_work_contacts",
                             "third_party_id"],
                "additionalProperties": False
            }
        },
        "interface_services": {
            "type": "object",
            "patternProperties": {
                "^.*$": {
                    "type": "object",
                    "patternProperties": {
                        "^.*$": {
                            "type": "object",
                            "patternProperties": {
                                "^.*$": {
                                    "type": "array",
                                    "items": {
                                        "type": "object",
                                        "properties": {
                                            "id": {"type": "integer"},
                                            "sid": {"type": "string"},
                                            "status": {"type": "string"},
                                            "name": {"type": "string"},
                                            "service_type": {"type": "string"},
                                            "contacts": {
                                                "type": "array",
                                                "items": {"type": "string"}
                                            },
                                            "planned_work_contacts": {
                                                "type": "array",
                                                "items": {"type": "string"}
                                            },
                                            "third_party_id": {"type": "string"}
                                        },
                                        "required": ["id", "sid", "status", "name", "service_type", "contacts",
                                                     "planned_work_contacts", "third_party_id"],
                                        "additionalProperties": False
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    },
    "required": ["circuit_hierarchy", "interface_services"],
    "additionalProperties": False
}


@routes.route("/all-data")
def get_everything():
    cache_key = "mic:impact:all-data"
    logger.debug(cache_key)
    r = common.get_current_redis()
    result = _ignore_cache_or_retrieve(request, cache_key, r)
    if not result:
        return Response(
            response='no data found',
            status=404,
            mimetype="text/html")
    return Response(result, mimetype='application/json')


@routes.route("/third-party-data")
def get_third_party_data():
    cache_key = "mic:impact:third-party-data"
    logger.debug(cache_key)
    r = common.get_current_redis()
    result = _ignore_cache_or_retrieve(request, cache_key, r)
    if not result:
        return Response(
            response='no data found',
            status=404,
            mimetype="text/html")
    return Response(result, mimetype='application/json')
